from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterBoolean,
  QgsFeatureRequest,
  QgsVectorLayer,
  QgsProcessingParameterCrs,
  QgsField,
  QgsProperty,
  QgsProcessingParameterField
  )

from qgis import processing


from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrFields, HrPostProcessor

class initelevationpoint(QgsProcessingAlgorithm):
  PARAMETERS = {                  
    "INPUT": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("initelevationpoint","Point layer"),
        "types": [QgsProcessing.TypeVectorPoint],
        "optional": True
      }
    },    
    "FIELD_USE_AS_HEIGHT": {
      "ui_func": QgsProcessingParameterField,
      "ui_args": {
        "optional": True,
        "description": QT_TRANSLATE_NOOP("initelevationpoint","Field indicating the height (if not specified, Z values of the points are used)"),
        "parentLayerParameterName": "INPUT",
        "defaultValue": None
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initelevationpoint","Target CRS (Cartesian coordinates)")
      }
    },
    "OVERWRITE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description" : QT_TRANSLATE_NOOP("initelevationpoint","Overwrite existing fields?"),
        "defaultValue": True
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initelevationpoint","Elevation points" )
      }
    }
  }
  
  FIELDS = {}
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)
    
    self.FIELDS["COMMON"] = HrFields.fromQgsFieldList([
      QgsField("PK", QVariant.Int, "int"),
      QgsField("HISTORY", QVariant.String, "string")      
    ]).setComments("note: common fields; ")
    
    self.FIELDS["REQUIRED"] = HrFields.fromQgsFieldList([
      QgsField("alti", QVariant.Double, "double", comment = "default: $z;"),
    ]).setComments("note: required fields; ")
    
      
  def initAlgorithm(self, config):
    (extent, target_crs) = self.API.getExtentAndCrsUsingCanvas()
    self.API.setDefaultValue("TARGET_CRS", target_crs.authid())
    self.API.initParameters()
    
  def processAlgorithm(self, parameters, context, feedback):   
    
    self.API.registerProcessingParameters(parameters, context, feedback)                              

    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    # check whether the target CRS is x-y coordinates
    self.API.checkCrsAsCartesian(target_crs)
    
    try:
      input_lyr = self.parameterAsSource(parameters, "INPUT", context).materialize(QgsFeatureRequest(), feedback)
    except:
      input_lyr = QgsVectorLayer(f"Polygon?crs={target_crs.authid()}", "temp", "memory")
      
    fld_z = self.parameterAsString(parameters, "FIELD_USE_AS_HEIGHT", context)
    if fld_z is not None and fld_z in input_lyr.fields().names():
      pnt_layer_with_height = processing.run(
        "native:setzvalue",
        {
          "INPUT": input_lyr,
          "Z_VALUE": QgsProperty.fromExpression(f'"{fld_z}"'),
          "OUTPUT": "memory:dem"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
    else:
      pnt_layer_with_height = input_lyr
    
    
    input_lyr_proj = processing.run(
      "native:reprojectlayer", 
      {
        "INPUT": pnt_layer_with_height,
        "TARGET_CRS": target_crs,
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]      
    input_lyr_fixed = context.getMapLayer(input_lyr_proj)
            
    
    
    try:
      self.FIELDS["EXISTING"] = HrFields.fromQgsFields(
        input_lyr_fixed.fields()
      ).setComments("note: existing fields; ", append = False)
    except:
      feedback.reportError(self.tr("The fields of input layer is invalid. The names may be duplicated (case-insensitive)."))
      raise Exception
    
    self.FIELDS["USE"] = self.FIELDS["COMMON"].concat(
      self.FIELDS["REQUIRED"]
    ).concat(
      self.FIELDS["EXISTING"],
      overwrite = not self.parameterAsBoolean(parameters, "OVERWRITE", context)
    )
    
    # feedback the results
    feedback.pushInfo(self.tr("The following fields are initialized: "))
    feedback.pushInfo(str(self.FIELDS["USE"]))    
    
    # define the feature sink
    new_fields = self.API.newFieldsWithHistory(self.FIELDS["USE"])
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      new_fields, 
      input_lyr_fixed.wkbType(), 
      input_lyr_fixed.sourceCrs()
    )
    
    self.CURRENT_PROCESS = self.API.parseCurrentProcess()
    n = input_lyr_fixed.featureCount()
    additional_attributes = {
      "PK": list(range(1, n+1)),
      "alti": []
    }
    
    for ft in input_lyr_fixed.getFeatures():
      additional_attributes["alti"].append(ft.geometry().get().z())
        
    self.API.addFeaturesWithHistoryToSink(
      sink = sink, vector_layer = input_lyr_fixed, fields = new_fields, 
      current_process = self.CURRENT_PROCESS,
      additional_attributes = additional_attributes
    )
    
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    
    return {"OUTPUT": dest_id}
  
  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Elevation point")

  def group(self):
    return self.tr("Initialize features")

  def groupId(self):
    return "initfeature"

  def createInstance(self):
    return initelevationpoint()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
