
from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsFeatureRequest,
  QgsVectorLayer,
  QgsField
  )

from qgis import processing
import os

from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrPostProcessor

class initroademissionfromtraffic(QgsProcessingAlgorithm):
  PARAMETERS = {  
    "INPUT": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("initroademissionfromtraffic","Road layer"),
        "types": [QgsProcessing.TypeVectorLine]
      },
      "n_mdl":"roadGeomPath",
      "save_layer_get_path": True
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initroademissionfromtraffic","Road" )
      }
    }
  }
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)
    self.NOISEMODELLING = {}
    
  def initAlgorithm(self, config):
    self.API.initParameters()
    
  def processAlgorithm(self, parameters, context, feedback):   
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    target_crs = self.API.parseCrs()
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.NOISEMODELLING = self.API.parseNoiseModellingArgs(
      {
        "GROOVY_SCRIPT": os.path.join(os.path.dirname(os.path.dirname(__file__)), "noisemodelling","hriskscript", "initroademissionfromtraffic.groovy"),
        "LW_ROADS": os.path.join("%nmtmp%", "LW_ROADS.geojson")
      },
      target_crs = target_crs,
      ui_parameters = self.PARAMETERS
    )
    
    # execute groovy script using wps_scripts
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])
    self.API.execNoiseModellingCmd(
      cmd      = self.NOISEMODELLING["CMD"],
      temp_dir = self.NOISEMODELLING["TEMP_DIR"]
    )
    
    # join to the source geom
    
    # first add level values to the receivers
    road_tr_layer = self.parameterAsSource(parameters, "INPUT", context).materialize(QgsFeatureRequest(), feedback)
    road_lw_layer = QgsVectorLayer(self.NOISEMODELLING["LW_ROADS"])
    
    lw_in_road_tr_layer = [fld for fld in road_tr_layer.fields().names() if fld[:3].lower() in ["lwd", "lwe","lwn"]]
    
    if len(lw_in_road_tr_layer) > 0:
      road_tr_layer = processing.run(
        "native:deletecolumn",
        {
          "INPUT": road_tr_layer,
          "COLUMN": lw_in_road_tr_layer,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]
      road_tr_layer = context.getMapLayer(road_tr_layer)
      
    road_joined = processing.run(
      "native:joinattributestable",
      {
        "INPUT": road_tr_layer,
        "FIELD": "PK",
        "INPUT_2": road_lw_layer,
        "FIELD_2": "PK",
        "FIELDS_TO_COPY": [fld for fld in road_lw_layer.fields().names() if fld not in ["PK"]],
        "METHOD": 0,
        "DISCARD_NONMATCHING": False,
        "PREFIX": "",
        "OUTPUT": "TEMPORARY_OUTPUT"
      },
      context = context,
      is_child_algorithm = True
    )["OUTPUT"]
    
    road_joined = context.getMapLayer(road_joined)
    
    if road_joined.featureCount() < 1:
      feedback.reportError(self.tr("The result of the NoiseModelling is empty."))
      raise Exception()
    
    new_fields = self.API.newFieldsWithHistory(road_joined.fields())
    # import the result    
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      new_fields, 
      road_joined.wkbType(), 
      road_joined.sourceCrs()
    )
    
    self.API.addFeaturesWithHistoryToSink(
      sink = sink, vector_layer = road_joined, fields = new_fields, 
      current_process = self.CURRENT_PROCESS
    )
    
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    
    return {"OUTPUT": dest_id}
    

  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Road emission calculated from traffic")

  def group(self):
    return self.tr("Initialize features")

  def groupId(self):
    return "initfeature"

  def createInstance(self):
    return initroademissionfromtraffic()
  
  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
