from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterBoolean,
  QgsFeatureRequest,
  QgsProcessingParameterCrs,
  QgsVectorLayer,
  QgsField
  )


from qgis import processing

from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrFields, HrPostProcessor

class initsource(QgsProcessingAlgorithm):
  PARAMETERS = {                  
    "INPUT": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("initsource","Source layer"),
        "types": [QgsProcessing.TypeVectorLine, QgsProcessing.TypeVectorPoint],
        "optional": True
      }
    },
    "USE_ZVALUE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description" : QT_TRANSLATE_NOOP("initsource","Use z values of the geometry?"),
        "defaultValue": False
      }
    },
    "TARGET_CRS": {
      "ui_func": QgsProcessingParameterCrs,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initsource","Target CRS (Cartesian coordinates)")
      }
    },
    "OVERWRITE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description" : QT_TRANSLATE_NOOP("initsource","Overwrite existing fields?"),
        "defaultValue": True
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("initsource","Source" )
      }
    }
  }
  
  FIELDS = {}
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)
    
    self.FIELDS["COMMON"] = HrFields.fromQgsFieldList([
      QgsField("PK", QVariant.Int, "int"),
      QgsField("HISTORY", QVariant.String, "string")      
    ]).setComments("note: common fields; ")
    
    self.FIELDS["REQUIRED"] = HrFields.fromQgsFieldList([
      QgsField("LWd63",   QVariant.Double, "double"), # Octave-band sound power level at 63 Hz (dB) (6-18h)
      QgsField("LWd125",  QVariant.Double, "double"), # Octave-band sound power level at 125 Hz (dB) (6-18h)
      QgsField("LWd250",  QVariant.Double, "double"), # Octave-band sound power level at 250 Hz (dB) (6-18h)
      QgsField("LWd500",  QVariant.Double, "double"), # Octave-band sound power level at 500 Hz (dB) (6-18h)
      QgsField("LWd1000", QVariant.Double, "double"), # Octave-band sound power level at 1000 Hz (dB) (6-18h)
      QgsField("LWd2000", QVariant.Double, "double"), # Octave-band sound power level at 2000 Hz (dB) (6-18h)
      QgsField("LWd4000", QVariant.Double, "double"), # Octave-band sound power level at 4000 Hz (dB) (6-18h)
      QgsField("LWd8000", QVariant.Double, "double"), # Octave-band sound power level at 8000 Hz (dB) (6-18h)
      QgsField("LWe63",   QVariant.Double, "double"), # Octave-band sound power level at 63 Hz (dB) (18-22h)
      QgsField("LWe125",  QVariant.Double, "double"), # Octave-band sound power level at 125 Hz (dB) (18-22h)
      QgsField("LWe250",  QVariant.Double, "double"), # Octave-band sound power level at 250 Hz (dB) (18-22h)
      QgsField("LWe500",  QVariant.Double, "double"), # Octave-band sound power level at 500 Hz (dB) (18-22h)
      QgsField("LWe1000", QVariant.Double, "double"), # Octave-band sound power level at 1000 Hz (dB) (18-22h)
      QgsField("LWe2000", QVariant.Double, "double"), # Octave-band sound power level at 2000 Hz (dB) (18-22h)
      QgsField("LWe4000", QVariant.Double, "double"), # Octave-band sound power level at 4000 Hz (dB) (18-22h)
      QgsField("LWe8000", QVariant.Double, "double"), # Octave-band sound power level at 8000 Hz (dB) (18-22h)
      QgsField("LWn63",   QVariant.Double, "double"), # Octave-band sound power level at 63 Hz (dB) (22-6h)
      QgsField("LWn125",  QVariant.Double, "double"), # Octave-band sound power level at 125 Hz (dB) (22-6h)
      QgsField("LWn250",  QVariant.Double, "double"), # Octave-band sound power level at 250 Hz (dB) (22-6h)
      QgsField("LWn500",  QVariant.Double, "double"), # Octave-band sound power level at 500 Hz (dB) (22-6h)
      QgsField("LWn1000", QVariant.Double, "double"), # Octave-band sound power level at 1000 Hz (dB) (22-6h)
      QgsField("LWn2000", QVariant.Double, "double"), # Octave-band sound power level at 2000 Hz (dB) (22-6h)
      QgsField("LWn4000", QVariant.Double, "double"), # Octave-band sound power level at 4000 Hz (dB) (22-6h)
      QgsField("LWn8000", QVariant.Double, "double"), # Octave-band sound power level at 8000 Hz (dB) (22-6h)  
    ])
  
  def initAlgorithm(self, config):
    (extent, target_crs) = self.API.getExtentAndCrsUsingCanvas()
    self.API.setDefaultValue("TARGET_CRS", target_crs.authid())
    self.API.initParameters()
    
  def processAlgorithm(self, parameters, context, feedback):
    self.API.registerProcessingParameters(parameters, context, feedback)
    target_crs = self.parameterAsCrs(parameters, "TARGET_CRS", context)
    # check whether the target CRS is x-y coordinates
    self.API.checkCrsAsCartesian(target_crs)

    try:
      input_lyr_proj = processing.run(
        "native:reprojectlayer", 
        {
          "INPUT": self.parameterAsSource(parameters, "INPUT", context).materialize(QgsFeatureRequest(), feedback),
          "TARGET_CRS": target_crs,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context,
        is_child_algorithm = True
      )["OUTPUT"]      
      input_lyr_proj = context.getMapLayer(input_lyr_proj)
            
    except:
      input_lyr_proj = QgsVectorLayer(f"Polygon?crs={target_crs.authid()}", "temp", "memory")
    
    if self.parameterAsBoolean(parameters, "USE_ZVALUE", context):
      test_geom = input_lyr_proj.getFeatures().__next__().geometry()
      if test_geom.get().nCoordinates() != 3:
        feedback.reportError(self.tr("The input layer has no Z values."))
        raise Exception(self.tr("The input layer has no Z values."))
      input_lyr_fixed = input_lyr_proj
    else:
      input_lyr_fixed = processing.run(
        "native:setzvalue",
        {
          "INPUT": input_lyr_proj,
          "Z_VALUE": 0.05,
          "OUTPUT": "TEMPORARY_OUTPUT"
        },
        context = context
        # is_child_algorithm = True
      )["OUTPUT"]
      
    try:
      self.FIELDS["EXISTING"] = HrFields.fromQgsFields(
        input_lyr_fixed.fields()
      ).setComments("note: existing fields; ", append = False)
    except:
      feedback.reportError(self.tr("The fields of input layer is invalid. The names may be duplicated (case-insensitive)."))
      raise Exception
    
    self.FIELDS["USE"] = self.FIELDS["COMMON"].concat(
      self.FIELDS["REQUIRED"]
    ).concat(
      self.FIELDS["EXISTING"],
      overwrite = not self.parameterAsBoolean(parameters, "OVERWRITE", context)
    )
       
    # feedback the results
    feedback.pushInfo(self.tr("The following fields are initialized: "))
    feedback.pushInfo(str(self.FIELDS["USE"]))    
    
    # define the feature sink
    new_fields = self.API.newFieldsWithHistory(self.FIELDS["USE"])
    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context, 
      new_fields, 
      input_lyr_fixed.wkbType(), 
      input_lyr_fixed.sourceCrs()
    )
    
    self.CURRENT_PROCESS = self.API.parseCurrentProcess()
    n = input_lyr_fixed.featureCount()
    additional_attributes = {
      "PK": list(range(1, n+1))
    }
    fields_new = self.FIELDS["COMMON"].concat(self.FIELDS["REQUIRED"])
    for fld, defval in zip(fields_new, HrFields.getDefaultValues(fields_new)):
      if fld.name().lower() != "pk" and defval is not None:
        additional_attributes[fld.name()] = [defval for i in range(n)]
        
    self.API.addFeaturesWithHistoryToSink(
      sink = sink, vector_layer = input_lyr_fixed, fields = new_fields, 
      current_process = self.CURRENT_PROCESS,
      additional_attributes = additional_attributes
    )
    
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
              
    return {"OUTPUT": dest_id}
  
  def name(self):
    return self.__class__.__name__

  def displayName(self):
    return self.tr("Sound source")

  def group(self):
    return self.tr("Initialize features")

  def groupId(self):
    return "initfeature"

  def createInstance(self):
    return initsource()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
