
from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterString,
  QgsProcessingParameterBoolean,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterNumber,
  QgsProcessingParameterFeatureSink,
  QgsFeatureRequest,
  QgsFeature,
  QgsGeometry,
  QgsVectorLayer,
  QgsWkbTypes
  )

import datetime
import os
from qgis import processing

from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrFields, HrPostProcessor

class noisefromtraffic(QgsProcessingAlgorithm):
  PARAMETERS = {
    "ROAD": {
      "crs_referrence": True, # this parameter is used as CRS referrence
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Road layer"),
        "types": [QgsProcessing.TypeVectorLine]
      },
      "n_mdl":"roadGeomPath",
      "save_layer_get_path": True
    },
    "BUILDING": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Building layer"),
        "types": [QgsProcessing.TypeVectorPolygon],
        "optional": True
      },
      "n_mdl": "buildingGeomPath",
      "save_layer_get_path": True
    },
    "RECEIVER": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Receiver layer"),
        "types": [QgsProcessing.TypeVectorPoint]
      },
      "n_mdl": "receiverGeomPath",
      "save_layer_get_path": True
    },
    "DEM": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Elevation point layer"),
        "types": [QgsProcessing.TypeVectorPoint],
        "optional": True
      },
      "n_mdl": "demGeomPath",
      "save_layer_get_path": True
    },
    "GROUND_ABS": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Ground absorption layer"),
        "types": [QgsProcessing.TypeVectorPolygon],
        "optional": True
      },
      "n_mdl":"groundAbsGeomPath",
      "save_layer_get_path": True
    },
  
    "MAX_SRC_DIST": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Max distance between source and receiver (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 100.0, "defaultValue": 500.0, "maxValue": 1000.0
      },
      "n_mdl": "confMaxSrcDist"
    },
    "MAX_REFL_DIST": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Max distance between source and reflection wall (m)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": 350.0, "maxValue": 1000.0
      },
      "n_mdl": "confMaxReflDist"
    },
    "REFL_ORDER": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Max number of reflections (times)"),
        "type": QgsProcessingParameterNumber.Integer,
        "minValue": 0, "defaultValue": 1, "maxValue": 5
      },
      "n_mdl": "confReflOrder"
    },
    "HUMIDITY": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Relative humidity (%)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": 70.0, "maxValue": 100.0
      },
      "n_mdl": "confHumidity"
    },
    "TEMPERATURE": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Temperature (°C)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": -40.0, "defaultValue": 15.0, "maxValue": 50.0
      },
      "n_mdl": "confTemperature"
    },
    "WALL_ALPHA": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Absorption coefficient at wall (0: fully reflective - 1: fully absorbent)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": 0.1, "maxValue": 1.0
      },
      "n_mdl": "paramWallAlpha"
    }, 
    "DIFF_VERTICAL": {
      "advanced": True,
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Diffraction at vertical edge"),
        "defaultValue": False
      },
      "n_mdl": "confDiffVertical"
    },
    "DIFF_HORIZONTAL": {
      "advanced": True,
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Diffraction at horizontal edge"),
        "defaultValue": True
      },
      "n_mdl": "confDiffHorizontal"
    },
    "FAV_OCURRENCE_DAY": {
      "advanced": True,
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Probability of occurrence of favorable condition (day)"),
        "defaultValue": "0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5"
      },
      "n_mdl": "confFavorableOccurrencesDay"
    },
    "FAV_OCURRENCE_EVENING": {
      "advanced": True,
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Probability of occurrence of favorable condition (evening)"),
        "defaultValue": "0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5"
      },
      "n_mdl": "confFavorableOccurrencesEvening"
    },
    "FAV_OCURRENCE_NIGHT": {
      "advanced": True,
      "ui_func": QgsProcessingParameterString,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Probability of occurrence of favorable condition (night)"),
        "defaultValue": "0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5"
      },
      "n_mdl": "confFavorableOccurrencesNight"
    },
    "THREAD_NUMBER": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Number of threads to calculate (0: all)"),
        "type": QgsProcessingParameterNumber.Integer,
        "minValue": 0, "defaultValue": 0, "maxValue": 20
      },
      "n_mdl": "confThreadNumber"
    },
    "SUBTRACTION_TO_CANCEL_REFLECTION": {
      "advanced": True,
      "ui_func": QgsProcessingParameterNumber,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Subtraction of sound level to cancel the reflection from the facade (dB)"),
        "type": QgsProcessingParameterNumber.Double,
        "minValue": 0.0, "defaultValue": 3.0, "maxValue": 10.0
      }
    },
    "ESTIMATE_BUILDING_LEVEL": {
      "advanced": True,
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Estimate sound levels of the buildings (only for facade receivers)"),
        "defaultValue": True
      }
    },
    "RUN_CALCULATION": {
      "advanced": True,
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Run sound-level calculation (or just export the calculation settings)?"),
        "defaultValue": True
      }      
    },
    "WPS_ARGS": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Arguments used for the calculation" )
      },
      "visibleByDefault": False
    },    
    "LDEN": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Lden" )
      },
      "visibleByDefault": True
    },    
    "LNIGHT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Lnight" )
      },
      "visibleByDefault": True
    },    
    "LDAY": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Lday" )
      },
      "visibleByDefault": False
    },    
    "LEVENING": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Levening" )
      },
      "visibleByDefault": False
    },    
    "BUILDING_WITH_LEVEL": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("noisefromtraffic","Building with facade level" )
      },
      "visibleByDefault": True
    }
  }
  
  BLDG_LEVEL_ARGS = {}
  
  def __init__(self):
    super().__init__()
    self.API = HrApi(self)  
    self.BLDG_LEVEL_ARGS = {
      "BUILDING_BID": "PK", # primary key of the buildings
      "RECEIVER_BID": "BUILD_PK", # primary key of the buildings, written in receiver layer
      "RECEIVER_RID": "PK", # primary key of the receivers
      "LEVEL_RID": "IDRECEIVER", #primary key of the receivers, written in level layer    
      "LEVEL_ASSIGN": ["LAEQ"], # which attribute(s) to be assined,
      "OVERWRITE": True
    }
  
  def initAlgorithm(self, config):
    self.API.initParameters()

  def processAlgorithm(self, parameters, context, feedback):
    
    self.API.registerProcessingParameters(parameters, context, feedback) 
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    target_crs = self.API.parseCrs()
    
    processing_results = {}
    group_name = "results_" + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    
    self.NOISEMODELLING = self.API.parseNoiseModellingArgs(
      {
        "GROOVY_SCRIPT": os.path.join("%grvhome%", "noisefromtraffic.groovy"),
        "RECEIVER": os.path.join("%nmtmp%", "RECEIVERS.geojson"),
        "LDAY":     os.path.join("%nmtmp%", "LDAY_GEOM.geojson"),
        "LEVENING": os.path.join("%nmtmp%", "LEVENING_GEOM.geojson"),
        "LNIGHT":   os.path.join("%nmtmp%", "LNIGHT_GEOM.geojson"),
        "LDEN":     os.path.join("%nmtmp%", "LDEN_GEOM.geojson"),
        "BUILDING_WITH_LEVEL": os.path.join("%nmtmp%", "BUILDING_WITH_LEVEL.geojson")
      },
      target_crs = target_crs,
      ui_parameters = self.PARAMETERS
    )
    
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])
    
    # output the wps args
    feedback.pushInfo(self.tr("Export the calculation settings."))
    
    wps_args = {
      "WPS_SCRIPT": self.NOISEMODELLING["CMD"].split()[0],
      "HISTORY": self.API.parseCurrentProcess(with_nm=True)
    }
    wps_args.update(self.NOISEMODELLING["WPS_ARGS"])
    
    wps_fields = self.API.newFieldsWithHistory(HrFields.fromDict(wps_args))
    (sink, dest_id) = self.parameterAsSink(
      parameters, "WPS_ARGS", context, 
      wps_fields, QgsWkbTypes.Polygon, target_crs
    )
    
    ft = QgsFeature(wps_fields)
    ft.setGeometry(QgsGeometry.fromRect(
      self.parameterAsSource(parameters, "RECEIVER", context).materialize(QgsFeatureRequest(), feedback).extent()
    ))
    for k, v in wps_args.items():
      ft[k] = v
    sink.addFeature(ft)
    
    processing_results["WPS_ARGS"] = dest_id    
    PostProcessors[dest_id] = HrPostProcessor(
      history = [self.CURRENT_PROCESS], coloring = "simple_bbox",
      group = group_name
    )
    
    if not self.parameterAsBoolean(parameters, "RUN_CALCULATION", context):    
      self.API.registerPostProcessAlgorithm(context)
      return processing_results
    
    self.API.execNoiseModellingCmd(
      cmd      = self.NOISEMODELLING["CMD"],
      temp_dir = self.NOISEMODELLING["TEMP_DIR"]
    )
    feedback.setProgress(100)
    
    
    # join sound levels to buildings, if specified
    if self.parameterAsBoolean(parameters, "ESTIMATE_BUILDING_LEVEL", context):
      feedback.pushInfo(self.tr("Estimating sound levels and health risks at buildings."))
      # set buildings and receivers layer
      bldg_layer = self.parameterAsSource(parameters, "BUILDING", context).materialize(QgsFeatureRequest(), feedback)
      rcv_layer = self.parameterAsSource(parameters, "RECEIVER", context).materialize(QgsFeatureRequest(), feedback)
      level_subtraction = self.parameterAsDouble(parameters, "SUBTRACTION_TO_CANCEL_REFLECTION", context)
      
      try:
        for noise_result in ["LDAY","LEVENING", "LNIGHT", "LDEN"]:
          feedback.pushInfo("Assigning " + noise_result + " values to the buildings.")
          args = {
            **self.BLDG_LEVEL_ARGS,
            **{
              "BUILDING": bldg_layer,
              "RECEIVER": rcv_layer,
              "LEVEL": self.NOISEMODELLING[noise_result],
              "LEVEL_PREFIX": noise_result + "_",
              "SUBTRACTION_TO_CANCEL_REFLECTION": level_subtraction,
              "OUTPUT": "TEMPORARY_OUTPUT"
            }
          }
          bldg_layer = processing.run(
            "hrisk:estimatelevelofbuilding",
            args, context = context, is_child_algorithm = True
          )["OUTPUT"]
          
        bldg_layer = context.getMapLayer(bldg_layer)  
        bldg_fields = self.API.newFieldsWithHistory(bldg_layer.fields())
        (sink, dest_id) = self.parameterAsSink(
          parameters, "BUILDING_WITH_LEVEL", context, 
          bldg_fields, bldg_layer.wkbType(), target_crs
        )
        self.API.addFeaturesWithHistoryToSink(
          sink, bldg_layer, bldg_fields, self.CURRENT_PROCESS
        )
        
        processing_results["BUILDING_WITH_LEVEL"] = dest_id
        PostProcessors[dest_id] = HrPostProcessor(
          history = [self.CURRENT_PROCESS], 
          coloring = "graduated_symbol", attribute = "LDEN_LAEQ_maximum", 
          group = group_name
        )
        
      except:
        feedback.pushWarning(
          self.tr("Warning: Sound levels are not assigned to the buildings. Inputs may be wrong:") + str(args)
          )
        
      
    # output sound-level results
    feedback.pushInfo(self.tr("Results of NoiseModelling are being loaded."))
    layers_to_output_dict = {
      k: v for k, v in self.NOISEMODELLING.items()
      if k in ["LDAY","LEVENING", "LNIGHT", "LDEN", "BUILDING_WITH_LEVEL"]
    }
    for file_name, file_path in layers_to_output_dict.items():
      if os.path.exists(file_path):
        result_layer = QgsVectorLayer(file_path, file_name)
        new_fields = self.API.newFieldsWithHistory(result_layer.fields())
        (sink, dest_id) = self.parameterAsSink(
          parameters, file_name, context,
          new_fields, result_layer.wkbType(), target_crs
        )
        self.API.addFeaturesWithHistoryToSink(
          sink, result_layer, new_fields, self.CURRENT_PROCESS
        )
        
        processing_results[file_name] = dest_id
        
        PostProcessors[dest_id] = HrPostProcessor(
          history = [self.CURRENT_PROCESS],
          group = group_name,            
          coloring = "graduated_symbol", attribute = "LAEQ", 
          theme = "Default"
        )
    
    self.API.registerPostProcessAlgorithm(context)
    return processing_results
  
  def name(self):
    return self.__class__.__name__
  
  def displayName(self):
    return self.tr("Prediction from traffic")

  def group(self):
    return self.tr("Predict sound level")

  def groupId(self):
    return "soundlevel"

  def createInstance(self):
    return noisefromtraffic()


  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
