from qgis.PyQt.QtCore import (QT_TRANSLATE_NOOP, QVariant, QCoreApplication)
from qgis.core import (
  QgsProcessing,
  QgsProcessingAlgorithm,
  QgsProcessingParameterFeatureSource,
  QgsProcessingParameterFeatureSink,
  QgsProcessingParameterBoolean,
  QgsFeatureRequest,
  QgsProcessingParameterField,
  QgsFeature,
  QgsField,
  QgsProcessingParameterNumber
  )

from ..hriskvar import PostProcessors
from ..hriskapi import HrApi, HrFields, HrPostProcessor

class estimateriskofbuilding(QgsProcessingAlgorithm):
  PARAMETERS = {
    "BUILDING": {
      "ui_func": QgsProcessingParameterFeatureSource,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Building layer"),
        "types": [QgsProcessing.TypeVectorPolygon]
      }
    },
    "LDEN": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Lden field"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "LDEN_LAEQ_Maximum"
      }
    },
    "LNIGHT": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Lnight field"),
        "parentLayerParameterName": "BUILDING",
        "defaultValue": "LNIGHT_LAEQ_Maximum"
      }
    },
    "POP": {
      "ui_func": QgsProcessingParameterField,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Population field"),
        "parentLayerParameterName": "BUILDING"
      }
    },
    "BASERISK_IHD_INC": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Base-line risk of IHD incident, per 100000"),
        "type": QgsProcessingParameterNumber.Double,
        "defaultValue": 0.0
      }
    },
    "BASERISK_IHD_DEATH": {
      "ui_func": QgsProcessingParameterNumber,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Base-line risk of IHD death, per 100000"),
        "type": QgsProcessingParameterNumber.Double,
        "defaultValue": 129.2012 # from WHO-GHE 2019
      }
    },
    "OVERWRITE": {
      "ui_func": QgsProcessingParameterBoolean,
      "ui_args":{
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Overwrite fields if they already exist?"),
        "defaultValue": True
      }
    },
    "OUTPUT": {
      "ui_func": QgsProcessingParameterFeatureSink,
      "ui_args": {
        "description": QT_TRANSLATE_NOOP("estimateriskofbuilding","Building with health risks")
      }
    }
  }
  
  FIELDS_ROAD = HrFields.fromQgsFieldList([
    QgsField(
      "relRiskIHD", QVariant.Double,
      comment = "formula: 1.08 ** ((ft['{LDEN}'] - 53.0) / 10) if ft['{LDEN}'] > 53.0 else 1.0; "
    ),
    QgsField(
      "probHA", QVariant.Double,
      comment = "formula: 0.78927 - 0.031162 * ft['{LDEN}'] + 0.000342 * ft['{LDEN}'] ** 2 if ft['{LDEN}'] > 40.0 else 0.0; "
    ),
    QgsField(
    "probHSD", QVariant.Double,
      comment = "formula: 0.194312 - 0.009336 * ft['{LNIGHT}'] + 0.000126 * ft['{LNIGHT}'] ** 2 if ft['{LNIGHT}'] > 40.0 else 0.0; "
    ),
    QgsField(
      "nHA", QVariant.Double,
      comment = "formula: ft_new['probHA'] * ft['{POP}'] if ft['{POP}'] != None else 0; "
    ),
    QgsField(
      "nHSD", QVariant.Double,
      comment = "formula: ft_new['probHSD'] * ft['{POP}'] if ft['{POP}'] != None else 0; "
    ),
    QgsField(
      "nIncIHD", QVariant.Double,
      comment = "formula: (ft_new['relRiskIHD'] - 1) * {INC_BR} * ft['{POP}'] if ft['{POP}'] != None else 0; "
    ),
    QgsField(
      "nDeathIHD", QVariant.Double,
      comment = "formula: (ft_new['relRiskIHD'] - 1) * {DEATH_BR} * ft['{POP}'] if ft['{POP}'] != None else 0; "
    )
  ])
  
  def __init__(self) -> None:
    super().__init__()
    self.API = HrApi(self)
  
  
  def initAlgorithm(self, config):
    self.API.initParameters()

  def processAlgorithm(self, parameters, context, feedback):   
    
    self.API.registerProcessingParameters(parameters, context, feedback)
    self.CURRENT_PROCESS = self.API.parseCurrentProcess(with_nm=True)
    
    
    bldg_layer = self.parameterAsSource(parameters, "BUILDING", context).materialize(QgsFeatureRequest(), feedback)
    bldg_fields = bldg_layer.fields()
    
    flds = []
    for fld in self.FIELDS_ROAD:
      fml = fld.comment().format(
        INC_BR = self.parameterAsDouble(parameters, "BASERISK_IHD_INC", context) / 100000.0,
        DEATH_BR = self.parameterAsDouble(parameters, "BASERISK_IHD_DEATH", context) / 100000.0,
        POP = self.parameterAsString(parameters, "POP", context),
        LDEN = self.parameterAsString(parameters, "LDEN", context),
        LNIGHT = self.parameterAsString(parameters, "LNIGHT", context)
      )
      
      fld.setComment(fml)
      flds.append(fld)
    
    self.FIELDS_ROAD = HrFields.fromQgsFieldList(flds)
    
    formulae = HrFields.getSpecifiedComments(self.FIELDS_ROAD, "formula")
    
    for existing_field_name in bldg_fields.names():
      if existing_field_name.lower() in map(lambda x: x.lower(), self.FIELDS_ROAD.names()):
        if self.parameterAsBoolean(parameters, "OVERWRITE", context):
          bldg_fields.remove(bldg_fields.indexOf(existing_field_name))
        else:
          feedback.reportError(self.tr("The field name is duplicated:") + existing_field_name)
          raise Exception(self.tr("The field name is duplicated"))
        
    for fld in self.FIELDS_ROAD:
      bldg_fields.append(fld)
    
    new_fields = self.API.newFieldsWithHistory(bldg_fields)

    (sink, dest_id) = self.parameterAsSink(
      parameters, "OUTPUT", context,
      new_fields, bldg_layer.wkbType(), bldg_layer.crs()
    )
    
    for ft in bldg_layer.getFeatures():
      
      ft_new = QgsFeature(bldg_fields)
      ft_new.setGeometry(ft.geometry())
      for field_name in ft.fields().names():        
        if field_name == "HISTORY":
          if len(ft["HISTORY"]) > 0:
            ft_new["HISTORY"] = ft["HISTORY"] + ";" + self.CURRENT_PROCESS
          else:
            ft_new["HISTORY"] = self.CURRENT_PROCESS
        elif field_name in bldg_fields.names():
          ft_new[field_name] = ft[field_name]
      
      for fld, fml in zip(self.FIELDS_ROAD, formulae):
        if fml != None:
          ft_new[fld.name()] = eval(fml)
          
      sink.addFeature(ft_new)
    
    PostProcessors[dest_id] = HrPostProcessor(history = [self.CURRENT_PROCESS])
    self.API.registerPostProcessAlgorithm(context)
          
    return {"OUTPUT": dest_id}

  def name(self):
    return self.__class__.__name__  

  def displayName(self):
    return self.tr("Estimate health risks of buildings")

  def group(self):
    return self.tr("Evaluate health risk")

  def groupId(self):
    return "healthrisk"
  
  def createInstance(self):
    return estimateriskofbuilding()

  # placing here is necessary, when employing pylupdate
  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
